//========================================================================
//
// LTKConfig.h
//
// Copyright 1996 Derek B. Noonburg
//
//========================================================================

#ifndef LTKCONFIG_H
#define LTKCONFIG_H

// default colors
#define LTK_FOREGROUND "black"
#define LTK_BACKGROUND "gray70"

// default window font
#define LTK_WIN_FONT "-*-helvetica-medium-r-normal-*-14-*-*-*-*-*-*-*"

// default font that every server should have
#define LTK_DEF_FONT "fixed"

#endif
